@@ -1,5 +1,5 @@
 package PAR;
-$PAR::VERSION = '1.011';
+$PAR::VERSION = '1.012';
 
 use 5.006;
 use strict;
@@ -18,6 +18,7 @@ BEGIN {
             File::Glob
             File::Spec
             File::Temp
+            Fcntl
             LWP::Simple
             PAR::Heavy
         /;
@@ -309,6 +310,9 @@ shared object file..."
 
 =cut
 
+use Fcntl ':flock';
+use Archive::Zip qw( :ERROR_CODES ); 
+
 use vars qw(@PAR_INC);              # explicitly stated PAR library files (preferred)
 use vars qw(@PAR_INC_LAST);         # explicitly stated PAR library files (fallback)
 use vars qw(%PAR_INC);              # sets {$par}{$file} for require'd modules
@@ -595,15 +599,16 @@ sub _first_member_matching {
 
 sub _run_member_from_par {
     my $member = shift;
-    my ($fh, $is_new, $filename) = _tempfile($member->crc32String . ".pl");
-
-    if ($is_new) {
-        my $file = $member->fileName;
-        print $fh "package main;\n",
-                  "#line 1 \"$file\"\n";
-        $member->extractToFileHandle($fh);
-        seek ($fh, 0, 0);
-    }
+    my (undef, $filename) = _tempfile(
+        sub {
+            my $fh = shift;
+            my $file = $member->fileName;
+            print $fh "package main;\n",
+                      "#line 1 \"$file\"\n";
+            $member->extractToFileHandle($fh) == AZ_OK
+                or die "Can't extract $file: $!";
+        },
+        $member->crc32String . ".pl");
 
     $ENV{PAR_0} = $filename; # for Pod::Usage
     { do $filename;
@@ -615,15 +620,16 @@ sub _run_member_from_par {
 
 sub _run_member {
     my $member = shift;
-    my ($fh, $is_new, $filename) = _tempfile($member->crc32String . ".pl");
-
-    if ($is_new) {
-        my $file = $member->fileName;
-        print $fh "package main;\n",
-                  "#line 1 \"$file\"\n";
-        $member->extractToFileHandle($fh);
-        seek ($fh, 0, 0);
-    }
+    my ($fh, $filename) = _tempfile(
+        sub {
+            my $fh = shift;
+            my $file = $member->fileName;
+            print $fh "package main;\n",
+                      "#line 1 \"$file\"\n";
+            $member->extractToFileHandle($fh) == AZ_OK
+                or die "Can't extract $file: $!";
+        },
+        $member->crc32String . ".pl");
 
     # NOTE: Perl 5.14.x will print the infamous warning
     # "Use of uninitialized value in do "file" at .../PAR.pm line 636" 
@@ -675,12 +681,20 @@ sub _extract_inc {
     my $is_handle = ref($file_or_azip_handle) && $file_or_azip_handle->isa('Archive::Zip::Archive');
 
     require File::Spec;
+
     my $inc = File::Spec->catdir($PAR::SetupTemp::PARTemp, "inc");
+    my $inc_lock = "$inc.lock";
+
     my $canary = File::Spec->catfile($PAR::SetupTemp::PARTemp, $PAR::SetupTemp::Canary);
 
-    if (!-d $inc || !-e $canary) {
-        for (1 .. 10) { mkdir("$inc.lock", 0755) and last; sleep 1 }
-        
+    # acquire the "wanna extract inc" lock
+    open my $lock, ">", $inc_lock or die qq[can't open "$inc_lock": $!];
+    flock($lock, LOCK_EX);
+
+    unless (-d $inc && -e $canary)
+    {
+        mkdir($inc, 0755);
+
         undef $@;
         if (!$is_handle) {
           # First try to unzip the *fast* way.
@@ -697,8 +711,8 @@ sub _extract_inc {
           }
         }
 
-        # either failed to load Archive::Unzip::Burst or got an A::Zip handle
-        # fallback to slow way.
+        # either failed to load Archive::Unzip::Burst or got 
+        # an Archive::Zip handle: fallback to slow way.
         if ($is_handle || $@) {
           my $zip;
           if (!$is_handle) {
@@ -708,15 +722,13 @@ sub _extract_inc {
             bless($fh, 'IO::File');
 
             $zip = Archive::Zip->new;
-            ( $zip->readFromFileHandle($fh, $file_or_azip_handle) == Archive::Zip::AZ_OK() )
+            $zip->readFromFileHandle($fh, $file_or_azip_handle) == AZ_OK
                 or die "Read '$file_or_azip_handle' error: $!";
           }
           else {
             $zip = $file_or_azip_handle;
           }
 
-          mkdir($inc) if not -d $inc;
-
           for ( $zip->memberNames() ) {
               s{^/}{};
               my $outfile =  File::Spec->catfile($inc, $_);
@@ -728,7 +740,9 @@ sub _extract_inc {
               utime(undef, undef, $outfile);
           }
         }
-        
+
+        $ArchivesExtracted{$is_handle ? $file_or_azip_handle->fileName() : $file_or_azip_handle} = $inc;
+    
         # touch (and back-date) canary file
         open my $fh, ">", $canary; 
         print $fh <<'...';
@@ -739,12 +753,12 @@ mechanism (probably based on file modification times).
         close $fh;
         my $dateback = time() - $PAR::SetupTemp::CanaryDateBack;
         utime($dateback, $dateback, $canary);
-
-        rmdir("$inc.lock");
-
-        $ArchivesExtracted{$is_handle ? $file_or_azip_handle->fileName() : $file_or_azip_handle} = $inc;
     }
 
+    # release the "wanna extract inc" lock
+    flock($lock, LOCK_UN);
+    close $lock;
+
     # add the freshly extracted directories to @INC,
     # but make sure there's no duplicates
     my %inc_exists = map { ($_, 1) } @INC;
@@ -961,9 +975,10 @@ sub unpar {
         }
         # Got the .par as a string. (reference to scalar, of course)
         elsif (ref($par) eq 'SCALAR') {
-            my ($fh) = _tempfile();
-            print $fh $$par;
-            $par = $fh;
+            ($par, undef) = _tempfile(sub {
+                    my $fh = shift;
+                    print $fh $$par;
+                });
         }
         # If the par is not a valid .par file name and we're being strict
         # about this, then also check whether "$par.par" exists
@@ -989,7 +1004,7 @@ sub unpar {
         Archive::Zip::setErrorHandler(sub {});
         my $rv = $zip->readFromFileHandle($par, @file);
         Archive::Zip::setErrorHandler(undef);
-        return unless $rv == Archive::Zip::AZ_OK();
+        return unless $rv == AZ_OK;
 
         push @LibCache, $zip;
         $LibCache{$_[0]} = $zip;
@@ -1025,11 +1040,13 @@ sub unpar {
                         ^([^/]+)$
                     };
                 my $extract_name = $1;
-                my $dest_name =
-                    File::Spec->catfile($ENV{PAR_TEMP}, $extract_name);
+                my $dest_name = File::Spec->catfile($ENV{PAR_TEMP}, $extract_name);
                 # but don't extract it if we've already got one
-                $member->extractToFileNamed($dest_name)
-                    unless(-e $dest_name);
+                unless (-e $dest_name)
+                {
+                    $member->extractToFileNamed($dest_name) == AZ_OK
+                        or die "Can't extract $member_name: $!";
+                }
             }
         }
 
@@ -1054,7 +1071,6 @@ sub unpar {
                $ENV{$key} = $tempdir;
            }
        }
-    
     }
 
     $LastAccessedPAR = $zip;
@@ -1072,51 +1088,60 @@ sub unpar {
 
     return $member if $member_only;
 
-    my ($fh, $is_new);
-    ($fh, $is_new, $LastTempFile) = _tempfile($member->crc32String . ".pm");
-    die "Bad Things Happened..." unless $fh;
-
-    if ($is_new) {
-        $member->extractToFileHandle($fh);
-        seek ($fh, 0, 0);
-    }
+    my ($fh, $LastTempFile) = _tempfile(
+        sub { 
+            my $fh = shift; 
+            my $file = $member->fileName;
+            $member->extractToFileHandle($fh) == AZ_OK
+                or die "Can't extract $file: $!";
+        },
+        $member->crc32String . ".pm");
 
     return $fh;
 }
 
 sub _tempfile {
-    my ($fh, $filename);
-    if ($ENV{PAR_CLEAN} or !@_) {
+    my ($callback, $name) = @_;
+    
+    if ($ENV{PAR_CLEAN} or !defined $name) {
         require File::Temp;
 
         if (defined &File::Temp::tempfile) {
             # under Win32, the file is created with O_TEMPORARY,
             # and will be deleted by the C runtime; having File::Temp
             # delete it has the only effect of giving ugly warnings
-            ($fh, $filename) = File::Temp::tempfile(
+            my ($fh, $filename) = File::Temp::tempfile(
                 DIR     => $PAR::SetupTemp::PARTemp,
                 UNLINK  => ($^O ne 'MSWin32' and $^O !~ /hpux/),
             ) or die "Cannot create temporary file: $!";
             binmode($fh);
-            return ($fh, 1, $filename);
+            $callback->($fh);
+            seek($fh, 0, 0);
+            return ($fh, $filename);
         }
     }
 
     require File::Spec;
 
     # untainting tempfile path
-    local $_ = File::Spec->catfile( $PAR::SetupTemp::PARTemp, $_[0] );
-    /^(.+)$/ and $filename = $1;
+    my ($filename) = File::Spec->catfile($PAR::SetupTemp::PARTemp, $name) =~ /^(.+)$/;
 
-    if (-r $filename) {
-        open $fh, '<', $filename or die $!;
+    unless (-r $filename) {
+        my $tempname = "$filename.$$";
+
+        open my $fh, '>', $tempname or die $!;
         binmode($fh);
-        return ($fh, 0, $filename);
+        $callback->($fh);
+        close($fh);
+
+        # FIXME why?
+        rename($tempname, $filename) or unlink($tempname); 
     }
 
-    open $fh, '+>', $filename or die $!;
+    open my $fh, '<', $filename or die $!;
     binmode($fh);
-    return ($fh, 1, $filename);
+
+    return ($fh, $filename);
 }
 
 # Given an Archive::Zip object, this generates a hash of
